﻿using System;
using System.Diagnostics;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;

namespace SpellChecker {

    /// <summary>
    /// Adapter class for custom spelling adorners on TextBox controls.
    /// </summary>
    public class AderantWpfPlainTextBoxAdapter : Keyoti.RapidSpell.Wpf.WpfTextBoxAdapter {
        private readonly bool nullAdorner;

        /// <summary>
        /// Initializes a new instance of the <see cref="AderantWpfPlainTextBoxAdapter"/> class.
        /// </summary>
        public AderantWpfPlainTextBoxAdapter(TextBox textBox, bool createCutCopyPasteReplacement, bool nullAdorner) :
            base(textBox, createCutCopyPasteReplacement) {
            this.nullAdorner = nullAdorner;
        }


        /// <summary>
        /// Creates the error adorner.
        /// </summary>
        public override Keyoti.RapidSpell.Wpf.ErrorAdornerBase CreateErrorAdorner(int start, int end, Keyoti.RapidSpell.Wpf.UnderlineStyle u, Color c) {
            if (nullAdorner) {
                return new AderantNullAdorner(theWrappedTextBox as TextBox, start, end, this, u, c);
            }
            return new AderantPlainAdorner(theWrappedTextBox as TextBox, start, end, this, u, c);
        }
    }



    /// <summary>
    /// 
    /// </summary>
    public class AderantPlainAdorner : Keyoti.RapidSpell.Wpf.ErrorAdorner {

        /// <summary>
        /// 
        /// </summary>
        /// <param name="adornedElement"></param>
        /// <param name="start"></param>
        /// <param name="end"></param>
        /// <param name="iayt"></param>
        /// <param name="u"></param>
        /// <param name="c"></param>
        public AderantPlainAdorner(TextBox adornedElement, int start, int end, Keyoti.RapidSpell.Wpf.IAYTTextBox iayt, Keyoti.RapidSpell.Wpf.UnderlineStyle u, Color c)
            : base(adornedElement, start, end, iayt, u, c) {
            IsHitTestVisible = false;
            adornedElement.IsVisibleChanged += AdornedElementIsVisibleChanged;
            Unloaded += AderantPlainAdornerUnloaded;
        }

        private void AderantPlainAdornerUnloaded(object sender, RoutedEventArgs e) {
            AdornedElement.IsVisibleChanged -= AdornedElementIsVisibleChanged;
            Unloaded -= AderantPlainAdornerUnloaded;
        }

        private void AdornedElementIsVisibleChanged(object sender, DependencyPropertyChangedEventArgs e) {
            InvalidateVisual();
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="drawingContext"></param>
        protected override void OnRender(DrawingContext drawingContext) {
            if (!AdornedElement.IsVisible || !this.active) {//active is false when the word has been ignored.
                return;
            }

            drawingContext.PushClip(new RectangleGeometry(new Rect(0, 0, AdornedElement.RenderSize.Width, AdornedElement.RenderSize.Height)));

            int firstVisibleLineIndex = this.tb.GetFirstVisibleLineIndex();
            int lastVisibleLineIndex = this.tb.GetLastVisibleLineIndex();
            if (lastVisibleLineIndex == -1 || firstVisibleLineIndex == -1) {
                Debug.WriteLine("Render is Impossible because there are no visible lines.");
                OnRenderImpossible();
                return;
            }

            int firstCharacterLineIndex;
            int lastCharacterLineIndex;
            try {
                firstCharacterLineIndex = this.tb.GetLineIndexFromCharacterIndex(this.start);
                lastCharacterLineIndex = this.tb.GetLineIndexFromCharacterIndex(this.end);
                if (firstCharacterLineIndex == -1 || lastCharacterLineIndex == -1) {
                    Debug.WriteLine("Render is Impossible");
                    OnRenderImpossible();
                    return;
                }
                if (firstCharacterLineIndex < firstVisibleLineIndex - 1 || firstCharacterLineIndex > lastVisibleLineIndex + 1) {
                    return;
                }
            } catch (ArgumentOutOfRangeException) {
                OnRenderImpossible();
                return;
            }

            int tL = tb.Text.Length;
            if (start < tL && end <= tL) {
                Rect selectionStartPosScr = tb.GetRectFromCharacterIndex(start);

                if (selectionStartPosScr.IsEmpty) {
                    Debug.WriteLine("Render is Impossible");
                    OnRenderImpossible();
                    return;
                }

                Debug.WriteLine(string.Format("start: {0}, end: {1}, tb.Text: {2}", start, end, string.IsNullOrWhiteSpace(tb.Text) ? "Null" : tb.Text));

                SolidColorBrush highlightFillBrush = new SolidColorBrush { Color = Colors.Yellow, Opacity = 0.2 };
                Pen pen = new Pen(Brushes.Yellow, 1);

                for (int index = firstCharacterLineIndex; index <= lastCharacterLineIndex; index++) {
                    //Get the index of the first character on this line (either the start of the line or the first character of the error).
                    int firstCharIndex = index == firstCharacterLineIndex ? start : tb.GetCharacterIndexFromLineIndex(index);

                    //Get the rectangle representing the last character on this line (either the end of the line or the last character in the error).
                    Rect lastCharRect = index == lastCharacterLineIndex ? tb.GetRectFromCharacterIndex(end) : tb.GetRectFromCharacterIndex(firstCharIndex + tb.GetLineLength(index) - 1, true);

                    //Extend the last character rectangle to encompass the up to and including the first character.
                    lastCharRect.Union(tb.GetRectFromCharacterIndex(firstCharIndex));

                    //Draw the rectangle.
                    drawingContext.DrawRectangle(highlightFillBrush, pen, lastCharRect);
                }
                Debug.WriteLine("Render succeeded.");
            }
        }
    }

    /// <summary>
    /// Null adorner which does does no rendering, essentially allowing us to disable error decoration under certain circumstances.
    /// </summary>
    public class AderantNullAdorner : Keyoti.RapidSpell.Wpf.ErrorAdorner {
        /// <summary>
        /// 
        /// </summary>
        /// <param name="adornedElement"></param>
        /// <param name="start"></param>
        /// <param name="end"></param>
        /// <param name="iayt"></param>
        /// <param name="u"></param>
        /// <param name="c"></param>
        public AderantNullAdorner(TextBox adornedElement, int start, int end, Keyoti.RapidSpell.Wpf.IAYTTextBox iayt, Keyoti.RapidSpell.Wpf.UnderlineStyle u, Color c)
            : base(adornedElement, start, end, iayt, u, c) {
            IsHitTestVisible = false;
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="drawingContext"></param>
        protected override void OnRender(DrawingContext drawingContext) {
        }
    }

}
