﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;
using System.Linq;
using System.Threading;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Threading;

namespace SpellChecker {

    /// <summary>
    /// Advanced textbox control which supports expansion codes and spell checking.
    /// </summary>
    public class AdvancedTextBox : TextBox, IFormServiceConsumer {
        ScrollViewer scrollViewer;

        private static readonly RoutedCommand spellCheckDocumentCommand = new RoutedCommand("SpellCheckDocument", typeof(AdvancedTextBox));

        static AdvancedTextBox() {
            //DefaultStyleKeyProperty.OverrideMetadata(typeof(AdvancedTextBox), new FrameworkPropertyMetadata(typeof(AdvancedTextBox)));

            CommandManager.RegisterClassCommandBinding(typeof(AdvancedTextBox), new CommandBinding(
                spellCheckDocumentCommand,
                SpellCheckDocumentCommandExecuted,
                SpellCheckDocumentCommandCanExecute));
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="AdvancedTextBox"/> class.
        /// </summary>
        public AdvancedTextBox() {
            InputBindings.Add(new KeyBinding(spellCheckDocumentCommand, new KeyGesture(Key.F7)));
            CustomContextMenuItems = new MenuItemCollection();
        }

        /// <summary>
        /// Is called when a control template is applied.
        /// </summary>
        public override void OnApplyTemplate() {
            base.OnApplyTemplate();

            scrollViewer = GetTemplateChild("PART_ContentHost") as ScrollViewer;
            if (scrollViewer != null) {
                scrollViewer.ScrollChanged += OnScrollViewScrollChanged;
            }

            if (SpellCheckingAvailable) {
                EnableDisableSpellChecking(SpellCheckingEnabled);
            }

            ContextMenu = new ContextMenu();
        }

        #region Context Menu

        /// <summary>
        /// Gets or sets the merged context menu.
        /// </summary>
        public MenuItemCollection CustomContextMenuItems {
            get { return (MenuItemCollection)GetValue(CustomContextMenuItemsProperty); }
            set { SetValue(CustomContextMenuItemsProperty, value); }
        }

        /// <summary>
        /// Backing dependency property for the <see cref="CustomContextMenuItems"/> property.
        /// </summary>
        public static readonly DependencyProperty CustomContextMenuItemsProperty = DependencyProperty.Register(
            "CustomContextMenuItems",
            typeof(MenuItemCollection),
            typeof(AdvancedTextBox),
            new FrameworkPropertyMetadata(null));

        /// <summary>
        /// Creates a default context menu.
        /// </summary>
        private List<object> CreateContextMenu() {

            // get icon resources
            ResourceDictionary iconDictionary = new ResourceDictionary();

            // build the context menu
            List<object> contextMenu = new List<object> {
                new Separator()
            };

            if (SpellCheckingEnabled) {

                // add spell checker suggestions, if we have any.
                IFormServiceHost host = FormServiceController.GetSpellCheckServiceContainer(this);
                if (host != null) {
                    object[] suggestions = null;

                    try {
                        suggestions = host.RapidSpellAsYouType.GetSuggestions();
                    } catch (Exception) {
                        Trace.WriteLine("KEYOTI: Unable to retrieve spelling suggestions.");
                    }

                    if (suggestions != null) {

                        if (suggestions.Count() == 0) {
                            contextMenu.Add(new MenuItem {
                                Header = "NoSpellingSuggestions",
                                IsEnabled = false
                            });
                        } else {
                            suggestions.ToList().ForEach(s => {

                                HeaderedItemsControl item = s as HeaderedItemsControl;

                                if (item != null && item.Header is String && !(
                                    (String)item.Header == host.RapidSpellAsYouType.AddMenuText ||
                                    (String)item.Header == host.RapidSpellAsYouType.IgnoreAllMenuText)) {

                                    item.FontWeight = FontWeights.Bold;
                                }
                                contextMenu.Add(s);
                            });
                        }
                        contextMenu.Add(new Separator());
                    }
                }
                if (SpellCheckingDialogEnabled) {
                    contextMenu.Add(new MenuItem {
                        Command = spellCheckDocumentCommand,
                        CommandTarget = this,
                        Header = "Spellcheck..."
                    });
                }
            }
            return contextMenu;
        }

        /// <summary>
        /// Called whenever an unhandled <see cref="System.Windows.FrameworkElement.ContextMenuOpening"/> routed event reaches this class in its route. Implement this method to add class handling for this event.
        /// </summary>
        /// <param name="e">Arguments of the event.</param>
        protected override void OnContextMenuOpening(ContextMenuEventArgs e) {
            base.OnContextMenuOpening(e);

            if (ContextMenu == null) {
                return;
            }

            ContextMenu.Items.Clear();
            CreateContextMenu().ForEach(i => ContextMenu.Items.Add(i));
        }

        #endregion

        #region Custom Rendering

        /// <summary>
        /// Called when [scroll view scroll changed].
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="System.Windows.Controls.ScrollChangedEventArgs"/> instance containing the event data.</param>
        protected void OnScrollViewScrollChanged(object sender, ScrollChangedEventArgs e) {
            InvalidateVisual();
        }

        /// <summary>
        /// Is called when content in this editing control changes.
        /// </summary>
        /// <param name="e"></param>
        protected override void OnTextChanged(TextChangedEventArgs e) {
            base.OnTextChanged(e);
            InvalidateVisual();
        }

        ///// <summary>
        ///// When overridden in a derived class, participates in rendering operations that are directed by the layout system. The rendering instructions for this element are not used directly when this method is invoked, and are instead preserved for later asynchronous use by layout and drawing.
        ///// </summary>
        ///// <param name="drawingContext">The drawing instructions for a specific element. This context is provided to the layout system.</param>
        //protected override void OnRender(DrawingContext drawingContext) {

        //    RectangleGeometry backgroundGeometry = new RectangleGeometry(new Rect(2, 2, scrollViewer.ActualWidth + 2, ActualHeight - 3));
        //    drawingContext.DrawDrawing(new GeometryDrawing(Background, new Pen(Background, 1.00), backgroundGeometry));
            
        //    RectangleGeometry clipGeometry = new RectangleGeometry(new Rect(2, 2, ActualWidth, ActualHeight - 4));
        //    drawingContext.PushClip(clipGeometry);

        //}

        #endregion

        # region Spell Checking Support

        /// <summary>
        /// Identifies the <see cref="SpellCheckingAvailable"/> dependency property.
        /// </summary>
        public static readonly DependencyProperty SpellCheckingAvailableProperty = DependencyProperty.Register(
            "SpellCheckingAvailable",
            typeof(bool),
            typeof(AdvancedTextBox),
            new FrameworkPropertyMetadata(true, new PropertyChangedCallback(OnSpellCheckingAvailablePropertyChanged)));

        /// <summary>
        /// Identifies the <see cref="SpellCheckingEnabled"/> dependency property.
        /// </summary>
        public static readonly DependencyProperty SpellCheckingEnabledProperty = DependencyProperty.Register(
            "SpellCheckingEnabled",
            typeof(bool),
            typeof(AdvancedTextBox),
            new FrameworkPropertyMetadata(false, new PropertyChangedCallback(OnSpellCheckingEnabledPropertyChanged)));

        /// <summary>
        /// Identifies the <see cref="SpellCheckingDialogEnabled"/> dependency property.
        /// </summary>
        public static readonly DependencyProperty SpellCheckingDialogEnabledProperty = DependencyProperty.Register(
            "SpellCheckingDialogEnabled",
            typeof(bool),
            typeof(AdvancedTextBox),
            new FrameworkPropertyMetadata(true));

        /// <summary>
        /// Identifies the <see cref="IsCurrentSpellCheckTarget"/> dependency property.
        /// </summary>
        public static readonly DependencyProperty IsCurrentSpellCheckTargetProperty = DependencyProperty.Register(
            "IsCurrentSpellCheckTarget",
            typeof(bool),
            typeof(AdvancedTextBox),
            new FrameworkPropertyMetadata(false));

        private void EnableDisableSpellChecking(object state) {

            bool enable = (bool)state;

            if (enable && SpellCheckingAvailable) {
                FormServiceController.RegisterElement(this, FormLevelServiceType.SpellCheck);
            } else {
                FormServiceController.UnregisterElement(this, FormLevelServiceType.SpellCheck);
                InvalidateVisual();
            }

        }

        /// <summary>
        /// Event handler called when the <see cref="SpellCheckingEnabled"/> dependency property changes.
        /// </summary>
        private static void OnSpellCheckingEnabledPropertyChanged(DependencyObject d, DependencyPropertyChangedEventArgs e) {

            AdvancedTextBox textBox = d as AdvancedTextBox;
            bool enableSpellCheck = (bool)e.NewValue;

            if (textBox != null && textBox.IsLoaded) {
                textBox.Dispatcher.BeginInvoke(DispatcherPriority.Normal, new SendOrPostCallback(textBox.EnableDisableSpellChecking), enableSpellCheck);
            }
        }

        private static void OnSpellCheckingAvailablePropertyChanged(DependencyObject d, DependencyPropertyChangedEventArgs e) {
            AdvancedTextBox textBox = d as AdvancedTextBox;

            if (textBox != null && textBox.IsLoaded) {
                textBox.Dispatcher.BeginInvoke(DispatcherPriority.Normal, new SendOrPostCallback(textBox.EnableDisableSpellChecking), textBox.SpellCheckingEnabled);
            }
        }

        /// <summary>
        /// Indicates that this control has spell checking support available. This is a dependancy property.
        /// </summary>
        [SuppressMessage("Microsoft.Naming", "CA1702:CompoundWordsShouldBeCasedCorrectly", MessageId = "SpellChecking")]
        public bool SpellCheckingAvailable {
            get { return (bool)GetValue(SpellCheckingAvailableProperty); }
            set { SetValue(SpellCheckingAvailableProperty, value); }
        }

        /// <summary>
        /// Enables or Disables the ability to launch a spell checking dialog from the context menu of this control.  This is a dependency property.
        /// </summary>
        public bool SpellCheckingDialogEnabled {
            get { return (bool)GetValue(SpellCheckingDialogEnabledProperty); }
            set { SetValue(SpellCheckingDialogEnabledProperty, value); }
        }

        /// <summary>
        /// Enables or Disables the ADERANT spell checking feature for this control, if it is available. This is a dependancy property.
        /// </summary>
        [SuppressMessage("Microsoft.Naming", "CA1702:CompoundWordsShouldBeCasedCorrectly", MessageId = "SpellChecking")]
        public bool SpellCheckingEnabled {
            get { return (bool)GetValue(SpellCheckingEnabledProperty); }
            set { SetValue(SpellCheckingEnabledProperty, value); }
        }

        /// <summary>
        /// Indicates that this control is the current control being spell checked.
        /// </summary>
        /// <value></value>
        public bool IsCurrentSpellCheckTarget {
            get { return (bool)GetValue(IsCurrentSpellCheckTargetProperty); }
            set { SetValue(IsCurrentSpellCheckTargetProperty, value); }
        }

        /// <summary>
        /// Can execute handler for the spell check document command.
        /// </summary>
        private static void SpellCheckDocumentCommandCanExecute(object sender, CanExecuteRoutedEventArgs e) {
            AdvancedTextBox textBox = sender as AdvancedTextBox;
            if (textBox == null) {
                return;
            }
            e.CanExecute = textBox.SpellCheckingDialogEnabled;
        }

        /// <summary>
        /// Execute handler for the spell check document command.
        /// </summary>
        private static void SpellCheckDocumentCommandExecuted(object sender, ExecutedRoutedEventArgs e) {
            AdvancedTextBox textBox = sender as AdvancedTextBox;
            if (textBox == null) {
                return;
            }
            e.Handled = true;

            FormServiceController.PerformSpellCheck(textBox);
        }

        #endregion

        /// <summary>
        /// Service types the control is interested in.
        /// </summary>
        public FormLevelServiceType ServiceTypes {
            get {
                return FormLevelServiceType.FindReplace | FormLevelServiceType.SpellCheck | FormLevelServiceType.FlaggedWords;
            }
        }
    }

    /// <summary>
    /// A collection of menu items.
    /// </summary>
    public class MenuItemCollection : Collection<object> { };
}
