﻿using System;
using System.Collections;
using System.Diagnostics;
using System.IO;
using System.Text;
using Keyoti.RapidSpell.Wpf;
using Microsoft.Win32;

namespace SpellChecker {
    
    /// <summary>
    /// Custom dictionary for keyoti to support multiple MS-word dictionaries.
    /// </summary>
    public class CustomUserDictionary : UserDictionary {

        private readonly string defaultCustomDicPath = string.Empty;
        private readonly ArrayList wordList;
        private bool isValid = true;
        
        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="dictionaryId"></param>
        public CustomUserDictionary(String dictionaryId) : this(dictionaryId, 100000) {
            base.Location = StorageLocation.IsolatedStorage;
        }


        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="dictionaryId"></param>
        /// <param name="maxNumWords"></param>
        // ReSharper disable UnusedParameter.Local
        public CustomUserDictionary(string dictionaryId, int maxNumWords) {

            // ReSharper restore UnusedParameter.Local
            wordList = new ArrayList(1);

            //IsValid will be set to true when we find one custom dic file
            isValid = false;

            //Also need to add other dictionaries
            RegistryKey registryKey = Registry.CurrentUser.OpenSubKey(@"Software\Microsoft\Shared Tools\Proofing Tools\1.0\Custom Dictionaries");

            if (registryKey == null) {
                //Nothing specified. Falling back to custom dic.
                string path = GetDefaultLocationCustomDic("Custom.dic");

                if (File.Exists(path)) {
                    defaultCustomDicPath = path;
                    AddWordsFromFile(path);
                } 
                return;
            }

            //We found some custom dics. Add these
            string[] customdics = registryKey.GetValueNames();
            foreach (string customdicValueName in customdics) {
                int customdicNumber;

                if (int.TryParse(customdicValueName, out customdicNumber)) {
                    string path = string.Empty;
                    object value = registryKey.GetValue(customdicValueName);
                    if (value is string) {
                        path = value as string;
                        if (path.IndexOf(@"\") < 0) {
                            // this custom dic is in the normal location. Get the Path
                            path = GetDefaultLocationCustomDic(path);
                        }
                    }


                    if (customdicNumber == 1) {
                        //This is the primary/default custom dic
                        if (File.Exists(path)) {
                            defaultCustomDicPath = path;
                        } 
                    }

                    AddWordsFromFile(path);
                }
            }
        }

        /// <summary>
        /// Is this userdictionary valid?
        /// </summary>
        /// <returns></returns>
        public override bool IsValid() {
            //return true if the word list exists and is valid.
            return isValid;
        }

        /// <summary>
        /// Add the new word to the word custom.dic
        /// </summary>
        /// <param name="word"></param>
        /// <returns></returns>
        public override bool AddWord(String word) {

            if (wordList.Contains(word)) {
                return true;
            }
            
            wordList.Add(word);
            
            if (File.Exists(defaultCustomDicPath)) {

                // Try and find out what encoding the dictionary file has. This is to support different office dictionary versions.
                //  (By default, office 2007 is unicode, office 2003 is ISO8859-1/Latin1)
                Encoding encoding = GetFileEncoding(defaultCustomDicPath);

                // open the file and add the word
                try {
                    using (StreamWriter writer = new StreamWriter(defaultCustomDicPath, true, encoding)) {
                        writer.WriteLine(word);
                    }

                } catch (IOException) {
                    
                }


            } else {
                return false;
            }

            return true;
        }

        /// <summary>
        /// Read all words
        /// </summary>
        public override int ReadAll(ArrayList a) {
            //used by RapidSpell to read the words from the DB.
            a.Clear();
            a.AddRange(wordList);
            return a.Count;
        }

        private static string GetDefaultLocationCustomDic(string customDicName) {
            string appDataPath = Environment.ExpandEnvironmentVariables("%appdata%");

            string path = Path.Combine(appDataPath, "Microsoft", "UProof", customDicName);

            if (!File.Exists(path)) {
                // For the XP/Office 2003 case 
                path = Path.Combine(appDataPath, "Microsoft", "Proof", customDicName);
            }

            return path;
        }

        private void AddWordsFromFile(string path) {
            if (File.Exists(path)) {
                isValid = true;

                // Try and find out what encoding the dictionary file has. This is to support different office dictionary versions.
                //  (By default, office 2007 is unicode, office 2003 is ISO8859-1/Latin1)
                Encoding encoding = GetFileEncoding(path);

                try {
                    using (StreamReader reader = new StreamReader(path, encoding)) {
                        string line;
                        while ((line = reader.ReadLine()) != null) {
                            wordList.Add(line);
                        }
                        reader.Close();
                    }
                } catch (IOException) {
                    
                }

            } 
        }

        /// <summary>
        /// Return the encoding of a text file.  Return Encoding.Default if no Unicode BOM (byte order mark) is found.
        /// </summary>
        /// <param name="filename"></param>
        /// <returns>The encoding of the file.</returns>
        public static Encoding GetFileEncoding(string filename) {
            Encoding result = null;
            FileInfo fileInfo = new FileInfo(filename);
            FileStream fileStream = null;

            try {
                fileStream = fileInfo.OpenRead();

                Encoding[] unicodeEncodings = { Encoding.BigEndianUnicode, Encoding.Unicode, Encoding.UTF8 };

                for (int i = 0; result == null && i < unicodeEncodings.Length; i++) {
                    fileStream.Position = 0;

                    byte[] preamble = unicodeEncodings[i].GetPreamble();

                    bool preamblesAreEqual = true;

                    for (int j = 0; preamblesAreEqual && j < preamble.Length; j++) {
                        preamblesAreEqual = preamble[j] == fileStream.ReadByte();
                    }

                    if (preamblesAreEqual) {
                        result = unicodeEncodings[i];
                    }
                }
            } catch (Exception) {
                Trace.WriteLine("An exception occured while determining encoding of file.");
            } finally {
                if (fileStream != null) {
                    fileStream.Close();
                }
            }

            if (result == null) {
                result = Encoding.Default;
            }

            return result;
        }
    }
}
