﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using Keyoti.RapidSpell.Wpf;

namespace SpellChecker {
    /// <summary>
    /// Class to handle form level control registrations and requests from control for form level services.
    /// </summary>
    public class FormServiceController {

        private readonly Window host;
        private readonly List<AderantWpfPlainTextBoxAdapter> spellCheckingTargets = new List<AderantWpfPlainTextBoxAdapter>();
        private readonly Queue<ElementRegistrationQueueItem> registrationQueue = new Queue<ElementRegistrationQueueItem>();

        #region Routed Events

        /// <summary>
        /// An event raised when an element registers or unregisters for form level services.
        /// </summary>
        public static readonly RoutedEvent ElementRegisterationEvent = EventManager.RegisterRoutedEvent(
            "ElementRegistration",
            RoutingStrategy.Bubble,
            typeof(EventHandler<ElementRegistrationEventArgs>),
            typeof(FormServiceController));

        #endregion

        /// <summary>
        /// Initializes a new instance of the <see cref="FormServiceController"/> class.
        /// </summary>
        /// <param name="host">The host.</param>
        public FormServiceController(Window host) {
            this.host = host;
            this.host.Loaded += OnHostLoaded;
            this.host.Unloaded += OnHostUnloaded;
            host.AddHandler(ElementRegisterationEvent, new EventHandler<ElementRegistrationEventArgs>(OnElementRegistrationHandler));
        }

        void OnHostUnloaded(object sender, RoutedEventArgs e) {
            if (host != null) {
                host.Unloaded -= OnHostUnloaded;
                host.RemoveHandler(ElementRegisterationEvent, new EventHandler<ElementRegistrationEventArgs>(OnElementRegistrationHandler));

                IFormServiceHost formServiceHost = host as IFormServiceHost;
                if (formServiceHost != null) {
                    formServiceHost.RapidSpellDialog.SpellCheckFinished -= OnRapidSpellDialogSpellCheckFinished;
                }
            }

            if (spellCheckingTargets != null) spellCheckingTargets.Clear();
            if (registrationQueue != null) registrationQueue.Clear();

            e.Handled = true;
        }

        private void OnHostLoaded(object sender, RoutedEventArgs e) {
            host.Loaded -= OnHostLoaded;

            IFormServiceHost formServiceHost = host as IFormServiceHost;
            if (formServiceHost != null) {

                CustomUserDictionary dictionary = KeyotiConfigurator.CreateUserDictionary();

                KeyotiConfigurator.ConfigureRapidSpellAsYouType(formServiceHost.RapidSpellAsYouType, dictionary);
                KeyotiConfigurator.ConfigureRapidSpellDialog(formServiceHost.RapidSpellDialog, dictionary);
                formServiceHost.RapidSpellDialog.ShowOptionsButton = true;
                formServiceHost.RapidSpellDialog.OptionsEnabled = true;
                formServiceHost.RapidSpellDialog.Options.UserDictionary = dictionary;
                RapidSpellAYTDialogCoupler coupler = new RapidSpellAYTDialogCoupler();
                coupler.RapidSpellAsYouType = formServiceHost.RapidSpellAsYouType;
                coupler.RapidSpellDialog = formServiceHost.RapidSpellDialog;
                formServiceHost.RapidSpellDialog.SpellCheckFinished += OnRapidSpellDialogSpellCheckFinished;
            }

            ProcessRegistrationQueue();
        }

        private void FindSpellCheckingTargets(Action<AderantWpfPlainTextBoxAdapter> callback) {
            var elements = FindControlsForServiceType(host, FormLevelServiceType.SpellCheck);
            elements.ToList().ForEach(targetElement => callback(new AderantWpfPlainTextBoxAdapter((TextBox)targetElement, false, false)));
        }


        private void OnRapidSpellDialogSpellCheckFinished(object sender, Keyoti.RapidSpell.Wpf.Event.SpellCheckEventArgs e) {
            spellCheckingTargets.ForEach(targetElement => ((IFormServiceHost)host).RapidSpellDialog.RemoveTextComponent(targetElement));
        }

        /// <summary>
        /// Gets the spell check service container for the current visual tree element.
        /// </summary>
        public static IFormServiceHost GetSpellCheckServiceContainer(UIElement element) {
            return TryFindParent<Window>(element) as IFormServiceHost;
        }

        /// <summary>
        /// Forces to spell check everything. Used when for example the matter changes and you want to re-spell check.
        /// </summary>
        public static void ForceSpellCheckAll(UIElement element) {
            IFormServiceHost host = GetSpellCheckServiceContainer(element);
            if (host != null) {
                host.RapidSpellAsYouType.ForceCheckAll();
            }
        }

        #region Element registration convenience methods

        /// <summary>
        /// Registers a control for inclusion in form based spell checking.
        /// </summary>
        public static void RegisterElement(FrameworkElement element, FormLevelServiceType serviceTypes) {
            if (element == null) {
                throw new ArgumentNullException("element");
            }

            Trace.WriteLine(String.Format("Registering element {0} for form level services.", element.GetType()));
            element.RaiseEvent(new ElementRegistrationEventArgs { Action = ElementRegistrationAction.Register, ServiceTypes = serviceTypes });
        }

        /// <summary>
        /// Unregisters a control for inclusion in form based spell checking.
        /// </summary>
        public static void UnregisterElement(FrameworkElement element, FormLevelServiceType serviceTypes) {
            if (element == null) {
                throw new ArgumentNullException("element");
            }

            Trace.WriteLine(String.Format("Unregistering element {0} from form level services.", element.GetType()));
            element.RaiseEvent(new ElementRegistrationEventArgs { Action = ElementRegistrationAction.Unregister, ServiceTypes = serviceTypes });
        }

        #endregion

        #region Event Handlers

        internal void ProcessRegistrationQueue() {

            IFormServiceHost formServiceHost = host as IFormServiceHost;
            if (formServiceHost == null) {
                return;
            }

            while (registrationQueue.Count > 0) {

                ElementRegistrationQueueItem queueItem = registrationQueue.Dequeue();

                if (queueItem.Action == ElementRegistrationAction.Register) {

                    if (queueItem.ServiceTypes.HasFlag(FormLevelServiceType.SpellCheck)) {
                        var target = new AderantWpfPlainTextBoxAdapter((TextBox)queueItem.Element, false, false);
                        spellCheckingTargets.Add(target);

                        formServiceHost.RapidSpellAsYouType.AddTextComponent(target);
                    }
                }

                if (queueItem.Action == ElementRegistrationAction.Unregister) {

                    if (queueItem.ServiceTypes.HasFlag(FormLevelServiceType.SpellCheck)) {

                        var target = spellCheckingTargets.FirstOrDefault(t => t.WrappedTextBox.Equals(queueItem.Element));
                        if (target != null) {
                            spellCheckingTargets.Remove(target);
                            formServiceHost.RapidSpellAsYouType.RemoveTextComponent(target);
                        }

                    }

                }
            }
        }

        /// <summary>
        /// Event handler called when a control has registered for form level spell checking.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        public void OnElementRegistrationHandler(Object sender, ElementRegistrationEventArgs e) {
            Trace.WriteLine("Form level handler has received a registration event");

            UIElement target = e.OriginalSource as UIElement;
            if (target == null) {
                return;
            }

            registrationQueue.Enqueue(new ElementRegistrationQueueItem { Element = target, Action = e.Action, ServiceTypes = e.ServiceTypes });

            IFormServiceHost formServiceHost = host as IFormServiceHost;
            if (formServiceHost != null && formServiceHost.RapidSpellAsYouType != null) {
                ProcessRegistrationQueue();
            }
        }

        #endregion

        /// <summary>
        /// Performs the spell check.
        /// </summary>
        public static void PerformSpellCheck(UIElement element) {

            IFormServiceHost host = GetSpellCheckServiceContainer(element);
            if (host != null) {
                host.ServiceController.FindSpellCheckingTargets(target => host.RapidSpellDialog.AddTextComponent(target));
                host.RapidSpellDialog.Check();
            }
        }

        /// <summary>
        /// Force the spell check without using the dialog.
        /// </summary>
        public static void ForceSpellCheck(UIElement element) {

            IFormServiceHost host = GetSpellCheckServiceContainer(element);
            if (host != null) {
                host.ServiceController.FindSpellCheckingTargets(target => host.RapidSpellDialog.AddTextComponent(target));
                //TODO spell check only this element
            }
        }

        /// <summary>
        /// Walks the visual tree looking for controls which require the service type specified.
        /// </summary>
        private static IEnumerable<DependencyObject> FindControlsForServiceType(DependencyObject root, FormLevelServiceType serviceType) {
            return TryFindChildren<DependencyObject>(root, c => c is IFormServiceConsumer && ((IFormServiceConsumer)c).ServiceTypes.HasFlag(serviceType), true);
        }

        public static IEnumerable<T> TryFindChildren<T>(DependencyObject parent, Predicate<T> selector, bool handlePopup) where T : DependencyObject {
            if (parent != null) {
                for (int i = 0; i < VisualTreeHelper.GetChildrenCount(parent); i++) {
                    DependencyObject child = VisualTreeHelper.GetChild(parent, i);
                    if (child != null && child is T && selector((T)child)) {
                        yield return (T)child;
                    }

                    foreach (T childOfChild in TryFindChildren(child, selector, handlePopup)) {
                        yield return childOfChild;
                    }
                }
            }
        }

        /// <summary>
        /// Finds a parent of a given child item on the visual tree.
        /// </summary>
        /// <typeparam name="T">The type of the queried item.</typeparam>
        /// <param name="child">A direct or indirect child of the queried item.</param>
        /// <returns>
        /// The first parent item that matches the submitted type parameter. If not matching item can be found, a null reference is being returned.
        /// </returns>
        public static T TryFindParent<T>(DependencyObject child)
          where T : DependencyObject {

            // ensure a child element has been provided
            if (child == null)
                return null;

            // get parent item
            DependencyObject parentObject = VisualTreeHelper.GetParent(child);

            // the end of the visual tree has been reacheds
            if (parentObject == null)
                return null;

            // if the parent is a popup root then search from the placement target upwards as the popup has no visual parent
            if (parentObject.ToString().Equals("System.Windows.Controls.Primitives.PopupRoot"))
                parentObject = ((FrameworkElement)parentObject).Parent;
            
            // check if the parent matches the type we’re looking for
            T parent = parentObject as T;
            if (parent != null) {
                return parent;
            } 
            // use recursion to proceed with next level
            return TryFindParent<T>(parentObject);
        }
    }

    /// <summary>
    /// Available form level services.
    /// </summary>
    [Flags]
    public enum FormLevelServiceType {
        /// <summary>
        /// Spell checking.
        /// </summary>
        SpellCheck = 1,
        /// <summary>
        /// Find and replace.
        /// </summary>
        FindReplace = 2,
        /// <summary>
        /// Flagged words
        /// </summary>
        FlaggedWords = 3
    }

    #region Event Args

    /// <summary>
    /// Contains state information and event data associated with the event.
    /// </summary>
    public class ElementRegistrationEventArgs : RoutedEventArgs {

        /// <summary>
        /// Initializes a new instance of the <see cref="ElementRegistrationEventArgs"/> class.
        /// </summary>
        public ElementRegistrationEventArgs() {
            RoutedEvent = FormServiceController.ElementRegisterationEvent;
        }

        /// <summary>
        /// Gets or sets the action to perform.
        /// </summary>
        public ElementRegistrationAction Action { get; set; }

        /// <summary>
        /// Gets or sets the services which this registration applies to.
        /// </summary>
        public FormLevelServiceType ServiceTypes { get; set; }

    }

    #endregion
}
